# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module that defines the ADAPT-VQE algorithm framework. ADAPT-VQE is a
variational approach that builds an ansatz iteratively, until a convergence
criteria or a maximum number of cycles is reached. Each iteration ("cycle")
of ADAPT consists in drawing an operator from a pre-defined operator pool,
selecting the one that impacts the energy the most, growing the ansatz circuit
accordingly, and optimizing the variational parameters using VQE.

Ref:
    Grimsley, H.R., Economou, S.E., Barnes, E. et al.
    An adaptive variational algorithm for exact molecular simulations on a
    quantum computer.
    Nat Commun 10, 3007 (2019). https://doi.org/10.1038/s41467-019-10988-2.
"""

import math
import warnings

from scipy.optimize import minimize
from openfermion import commutator
from openfermion import FermionOperator as ofFermionOperator

from tangelo.toolboxes.operators.operators import FermionOperator, QubitOperator
from tangelo.toolboxes.ansatz_generator.adapt_ansatz import ADAPTAnsatz
from tangelo.toolboxes.qubit_mappings.mapping_transform import fermion_to_qubit_mapping
from tangelo.toolboxes.ansatz_generator._general_unitary_cc import uccgsd_generator as uccgsd_pool
from tangelo.toolboxes.operators import qubitop_to_qubitham
from tangelo.linq import Circuit, Gate
from tangelo.algorithms.variational.vqe_solver import VQESolver


class ADAPTSolver:
    """ADAPT VQE class. This is an iterative algorithm that uses VQE. Methods
    are defined to rank operators with respect to their influence on the total
    energy.

    Attributes:
        molecule (SecondQuantizedMolecule): The molecular system.
        tol (float): Maximum gradient allowed for a particular operator  before
            convergence.
        max_cycles (int): Maximum number of iterations for ADAPT.
        pool (func): Function that returns a list of FermionOperator. Each
            element represents excitation/operator that has an effect of the
            total energy.
        pool_args (dict) : The arguments for the pool function. Will be unpacked in
            function call as pool(**pool_args)
        qubit_mapping (str): One of the supported qubit mapping identifiers.
        qubit_hamiltonian (QubitOperator-like): Self-explanatory.
        up_then_down (bool): Spin orbitals ordering.
        n_spinorbitals (int): Self-explanatory.
        n_electrons (int): Self-explanatory.
        optimizer (func): Optimization function for VQE minimization.
        backend_options (dict): Backend options for the underlying VQE object.
        verbose (bool): Flag for verbosity of VQE.
        deflation_circuits (list[Circuit]): Deflation circuits to add an
            orthogonalization penalty with.
        deflation_coeff (float): The coefficient of the deflation.
        ref_state (array or Circuit): The reference configuration to use. Replaces HF state
     """

    def __init__(self, opt_dict):

        default_backend_options = {"target": None, "n_shots": None, "noise_model": None}
        default_options = {"molecule": None,
                           "tol": 1e-3, "max_cycles": 15,
                           "pool": uccgsd_pool,
                           "pool_args": None,
                           "qubit_mapping": "jw",
                           "qubit_hamiltonian": None,
                           "up_then_down": False,
                           "n_spinorbitals": None,
                           "n_electrons": None,
                           "spin": None,
                           "optimizer": self.LBFGSB_optimizer,
                           "backend_options": default_backend_options,
                           "verbose": False,
                           "ref_state": None,
                           "deflation_circuits": list(),
                           "deflation_coeff": 1}

        # Initialize with default values
        self.__dict__ = default_options
        # Overwrite default values with user-provided ones, if they correspond to a valid keyword
        for k, v in opt_dict.items():
            if k in default_options:
                setattr(self, k, v)
            else:
                # TODO Raise a warning instead, that variable will not be used unless user made mods to code
                raise KeyError(f"Keyword :: {k}, not available in {self.__class__.__name__}")

        # Raise error/warnings if input is not as expected. Only a single input
        # must be provided to avoid conflicts.
        if not (bool(self.molecule) ^ bool(self.qubit_hamiltonian)):
            raise ValueError(f"A molecule OR qubit Hamiltonian object must be provided when instantiating {self.__class__.__name__}.")

        self.ansatz = None
        self.converged = False
        self.iteration = 0
        self.energies = list()

        self.optimal_energy = None
        self.optimal_var_params = None
        self.optimal_circuit = None

    @property
    def operators(self):
        if self.ansatz is None:
            warnings.warn("operators: this attribute requires the 'build' method to be called first.")
            return None

        return self.ansatz.operators

    @property
    def ferm_operators(self):
        if self.ansatz is None:
            warnings.warn("ferm_operators: this attribute requires the 'build' method to be called first.")
            return None

        return self.ansatz.ferm_operators

    def build(self):
        """Builds the underlying objects required to run the ADAPT-VQE
        algorithm.
        """

        # Building molecule data with a pyscf molecule.
        if self.molecule:

            self.n_spinorbitals = self.molecule.n_active_sos
            self.n_electrons = self.molecule.n_active_electrons
            self.spin = self.molecule.spin

            # Compute qubit hamiltonian for the input molecular system
            self.qubit_hamiltonian = fermion_to_qubit_mapping(fermion_operator=self.molecule.fermionic_hamiltonian,
                                                              mapping=self.qubit_mapping,
                                                              n_spinorbitals=self.n_spinorbitals,
                                                              n_electrons=self.n_electrons,
                                                              up_then_down=self.up_then_down,
                                                              spin=self.spin)

        # Build / set ansatz circuit.
        ansatz_options = {"mapping": self.qubit_mapping, "up_then_down": self.up_then_down,
                          "reference_state": "HF" if self.ref_state is None else "zero"}
        self.ansatz = ADAPTAnsatz(self.n_spinorbitals, self.n_electrons, ansatz_options)

        # Build underlying VQE solver. Options remain consistent throughout the ADAPT cycles.
        self.vqe_options = {"qubit_hamiltonian": self.qubit_hamiltonian,
                            "ansatz": self.ansatz,
                            "optimizer": self.optimizer,
                            "backend_options": self.backend_options,
                            "deflation_circuits": self.deflation_circuits,
                            "deflation_coeff": self.deflation_coeff,
                            "ref_state": self.ref_state
                            }

        self.vqe_solver = VQESolver(self.vqe_options)
        self.vqe_solver.build()

        # If applicable, give vqe_solver access to molecule object
        if self.molecule:
            self.vqe_solver.molecule = self.molecule

        # Getting the pool of operators for the ansatz. If more functionalities
        # are added, this part must be modified and generalized.
        if self.pool_args is None:
            if self.pool == uccgsd_pool:
                self.pool_args = {"n_qubits": self.n_spinorbitals}
            else:
                raise KeyError('pool_args must be defined if using own pool function')

        # Check if pool function returns a QubitOperator or FermionOperator and populate variables
        pool_list = self.pool(**self.pool_args)

        # Only a qubit operator is provided with a FermionOperator pool.
        if not (self.n_spinorbitals and self.n_electrons and self.spin is not None):
            raise ValueError("Expecting the number of spin-orbitals (n_spinorbitals), "
                "the number of electrons (n_electrons) and the spin (spin) with "
                "a qubit_hamiltonian when working with a pool of fermion operators.")

        if isinstance(pool_list[0], QubitOperator):
            self.pool_type = 'qubit'
            self.pool_operators = pool_list
        elif isinstance(pool_list[0], (FermionOperator, ofFermionOperator)):
            self.pool_type = 'fermion'
            self.fermionic_operators = pool_list
            self.pool_operators = [fermion_to_qubit_mapping(fermion_operator=fi,
                                                            mapping=self.qubit_mapping,
                                                            n_spinorbitals=self.n_spinorbitals,
                                                            n_electrons=self.n_electrons,
                                                            up_then_down=self.up_then_down,
                                                            spin=self.spin) for fi in self.fermionic_operators]
        else:
            raise ValueError('pool function must return either QubitOperator or FermionOperator')

        # Cast all coefs to floats (rotations angles are real).
        for qubit_op in self.pool_operators:
            for term, coeff in qubit_op.terms.items():
                qubit_op.terms[term] = math.copysign(1., coeff.imag)

        # Getting commutators to compute gradients:
        # \frac{\partial E}{\partial \theta_n} = \langle \psi | [\hat{H}, A_n] | \psi \rangle
        self.pool_commutators = [commutator(self.qubit_hamiltonian, element) for element in self.pool_operators]

    def simulate(self):
        """Performs the ADAPT cycles. Each iteration, a VQE minimization is
        done.
        """

        params = self.vqe_solver.ansatz.var_params

        # Construction of the ansatz. self.max_cycles terms are added, unless
        # all operator gradients are less than self.tol.
        while self.iteration < self.max_cycles:
            self.iteration += 1
            if self.verbose:
                print(f"Iteration {self.iteration} of ADAPT-VQE.")

            full_circuit = (self.vqe_solver.ansatz.circuit if self.ref_state is None else
                            self.vqe_solver.reference_circuit + self.vqe_solver.ansatz.circuit)
            pool_select = self.rank_pool(self.pool_commutators, full_circuit,
                                         backend=self.vqe_solver.backend, tolerance=self.tol)

            # If pool selection returns an operator that changes the energy by
            # more than self.tol. Else, the loop is complete and the energy is
            # considered as converged.
            if pool_select > -1:

                # Adding a new operator + initializing its parameters to 0.
                # Previous parameters are kept as they were.
                params += [0.]
                if self.pool_type == 'fermion':
                    self.vqe_solver.ansatz.add_operator(self.pool_operators[pool_select], self.fermionic_operators[pool_select])
                else:
                    self.vqe_solver.ansatz.add_operator(self.pool_operators[pool_select])
                self.vqe_solver.initial_var_params = params

                # Performs a VQE simulation and append the energy to a list.
                # Also, forcing params to be a list to make it easier to append
                # new parameters. The behavior with a np.array is multiplication
                # with broadcasting (not wanted).
                self.vqe_solver.simulate()
                opt_energy = self.vqe_solver.optimal_energy
                params = list(self.vqe_solver.optimal_var_params)
                self.energies.append(opt_energy)
            else:
                self.converged = True
                break

        return self.energies[-1]

    def rank_pool(self, pool_commutators, circuit, backend, tolerance=1e-3):
        """Rank pool of operators with a specific circuit.

        Args:
            pool_commutators (QubitOperator): Commutator [H, operator] for each
                generator.
            circuit (tangelo.linq.Circuit): Circuit for measuring each commutator.
            backend (tangelo.linq.Simulator): Backend to compute expectation values.
            tolerance (float): Minimum value for gradient to be considered.

        Returns:
            int: Index of the operators with the highest gradient. If it is not
                bigger than tolerance, returns -1.
        """

        gradient = [abs(backend.get_expectation_value(element, circuit)) for element in pool_commutators]
        for deflate_circuit in self.deflation_circuits:
            for i, pool_op in enumerate(self.pool_operators):
                op_circuit = Circuit([Gate(op[1], op[0]) for tuple in pool_op.terms for op in tuple])
                pool_over = deflate_circuit.inverse() + op_circuit + circuit
                f_dict, _ = backend.simulate(pool_over)
                grad = f_dict.get("0"*self.vqe_solver.ansatz.circuit.width, 0)
                pool_over = deflate_circuit.inverse() + circuit
                f_dict, _ = backend.simulate(pool_over)
                gradient[i] += self.deflation_coeff * grad * f_dict.get("0"*self.vqe_solver.ansatz.circuit.width, 0)
        max_partial = max(gradient)

        if self.verbose:
            print(f"LARGEST PARTIAL DERIVATIVE: {max_partial :4E} \t[{gradient.index(max_partial)}]")

        return gradient.index(max_partial) if max_partial >= tolerance else -1

    def get_resources(self):
        """Returns resources currently used in underlying VQE."""

        return self.vqe_solver.get_resources()

    def LBFGSB_optimizer(self, func, var_params):
        """Default optimizer for ADAPT-VQE."""

        result = minimize(func, var_params, method="L-BFGS-B",
                          options={"disp": False, "maxiter": 100, "gtol": 1e-10, "iprint": -1})

        self.optimal_var_params = result.x
        self.optimal_energy = result.fun

        # Reconstructing the optimal circuit at the end of the ADAPT iterations
        # or when the algorithm has converged.
        if self.converged or self.iteration == self.max_cycles:
            self.ansatz.build_circuit(self.optimal_var_params)
            self.optimal_circuit = (self.vqe_solver.ansatz.circuit if self.ref_state is None else
                                    self.vqe_solver.reference_circuit + self.vqe_solver.ansatz.circuit)

        if self.verbose:
            print(f"VQESolver optimization results:")
            print(f"\tOptimal VQE energy: {result.fun}")
            print(f"\tOptimal VQE variational parameters: {result.x}")
            print(f"\tNumber of Iterations : {result.nit}")
            print(f"\tNumber of Function Evaluations : {result.nfev}")
            print(f"\tNumber of Gradient Evaluations : {result.njev}")

        return result.fun, result.x



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file provides utility functions to perform mathematical operations."""

import numpy as np


def bool_col_echelon(bool_array):
    """Function to transform a boolean array into its column echelon form
    (transpose of the row-echelon form).

    Args:
        bool_array (array of bool): Self-explanatory.
    """

    pivot = bool_array.shape[1] - 1

    # This is done on matrix where the number of rows > number of columns.
    active_rows = bool_array.shape[0] - bool_array.shape[1] - 1

    # Column by column, starting from the last one, perform gaussian elimination
    # on the columns.
    for row in range(active_rows, -1, -1):

        # Identify if there is a 1 in the selected row and columns.
        if bool_array[row, :pivot+1].max():
            indices = np.where(bool_array[row, :pivot+1])[0]

            # If there are more than one 1 in the selected part, perform
            # gaussian elimination (done with a XOR because the input is a
            # boolean array.
            if len(indices) > 1:
                for i in range(1, len(indices)):
                    bool_array[:, indices[i]] = np.logical_xor(bool_array[:, indices[i]],
                                                               bool_array[:, indices[0]])

            # If there is only one 1 in the selected part, the pivot is
            # decremented.
            if len(indices) > 0:
                bool_array[:, (indices[0], pivot)] = bool_array[:, (pivot, indices[0])]
                pivot -= 1

    return bool_array



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module implements a collection of functions related to the ILC ansatz:
1. Function to create the anticommuting set (ACS) of generators from the QCC DIS;
2. An efficient solver that performs Gaussian elimination over GF(2);
3. Function that computes the ILC parameters via matrix diagonalization.

Refs:
    1. R. A. Lang, I. G. Ryabinkin, and A. F. Izmaylov.
        arXiv:2002.05701v1, 2020, 1–10.
    2. R. A. Lang, I. G. Ryabinkin, and A. F. Izmaylov.
        J. Chem. Theory Comput. 2021, 17, 1, 66–78.
    3. Ç. K. Koç and S. N. Arachchige.
        J. Parallel Distrib. Comput., 1991, 13, 118–122.
"""

import warnings

import scipy
import numpy as np

from tangelo.toolboxes.operators.operators import QubitOperator
from tangelo.toolboxes.ansatz_generator._qubit_mf import get_op_expval


def construct_acs(dis, max_ilc_gens, n_qubits):
    """Driver function for constructing the anticommuting set of generators from
    the direct interaction set (DIS) of QCC generators.

    Args:
        dis (list of list): DIS of QCC generators.
        max_ilc_gens (int): maximum number of ILC generators allowed in the ansatz.
        n_qubits (int): number of qubits

    Returns:
        list of QubitOperator: the anticommuting set (ACS) of ILC generators
    """

    bad_sln_idxs, good_sln = [], False
    while not good_sln:
        gen_idxs, ilc_gens = [idx for idx in range(max_ilc_gens) if idx not in bad_sln_idxs], []
        n_gens = len(gen_idxs)
        ng2, ngnq = n_gens * (n_gens + 1) // 2, n_gens * n_qubits

        # a_mat --> A and z_vec --> z in Appendix A, Refs. 1 & 2.
        a_mat, z_vec, one_vec = np.zeros((ng2, ngnq)), np.zeros(ngnq), np.ones((ng2, 1))
        for idx, gen_idx in enumerate(gen_idxs):
            gen = dis[gen_idx]
            for term in gen.terms:
                for paulis in term:
                    p_idx, pauli = paulis
                    if 'X' in pauli or 'Y' in pauli:
                        z_vec[idx * n_qubits + p_idx] = 1.

        # Form the rectangular matrix-vector product A * z (Appendix A, Refs. 1 & 2).
        rowdx = 0
        for i in range(n_gens):
            a_mat[rowdx, i * n_qubits:(i+1) * n_qubits] = z_vec[i * n_qubits:(i+1) * n_qubits]
            rowdx += 1
            for j in range(i + 1, n_gens):
                a_mat[rowdx, i * n_qubits:(i+1) * n_qubits] = z_vec[j * n_qubits:(j+1) * n_qubits]
                a_mat[rowdx, j * n_qubits:(j+1) * n_qubits] = z_vec[i * n_qubits:(i+1) * n_qubits]
                rowdx += 1

        # Solve A * z = b --> here b = 1
        z_sln = gauss_elim_over_gf2(a_mat, b_vec=one_vec)

        # Check solution: odd # of Y ops, at least two flip indices, and mutually anticommutes
        for i in range(n_gens):
            n_flip, n_y, gen_idx, gen_tup = 0, 0, gen_idxs[i], tuple()
            for j in range(n_qubits):
                gen, idx = None, i * n_qubits + j
                if z_vec[idx] == 1.:
                    n_flip += 1
                    if z_sln[idx] == 0.:
                        gen = (j, 'X')
                    else:
                        gen = (j, 'Y')
                        n_y += 1
                else:
                    if z_sln[idx] == 1.:
                        gen = (j, 'Z')
                if gen:
                    gen_tup += (gen, )
            # check number of flip indices and number of Y Pauli ops
            if n_flip > 1 and n_y % 2 == 1:
                gen_i = QubitOperator(gen_tup, 1.)
                good_sln = True
                # check mutual anticommutativity of each new ILC generator with all the rest
                for gen_j in ilc_gens:
                    if gen_i * gen_j != -1. * gen_j * gen_i:
                        if gen_idx not in bad_sln_idxs:
                            bad_sln_idxs.append(gen_idx)
                            good_sln = False
            else:
                if gen_idx not in bad_sln_idxs:
                    bad_sln_idxs.append(gen_idx)
                    good_sln = False
            if good_sln:
                ilc_gens.append(gen_i)
    return ilc_gens


def gauss_elim_over_gf2(a_mat, b_vec=None):
    """Driver function that performs Gaussian elimination to solve A * z = b
    over the binary field where b is the known solution vector. This routine
    was adapted based on Ref. 3. All elements of a_mat and b_vec are assumed
    to be the integers 0 or 1.

    Args:
        a_mat (numpy array of int): rectangular matrix of dimension n x m that
            holds the action of A * z, where z is a column vector of dimension m x 1.
            No default.
        b_vec (numpy array of int): column vector of dimension n x 1 holding the
            initial solution of A * z. Default, np.zeros((n, 1)).

    Returns:
        numpy array of float: solution for the z vector of dimension (n, )
    """

    n_rows, n_cols = np.shape(a_mat)
    z_vals, z_sln, piv_idx = [], [-1] * n_cols, 0
    # check that b_vec was properly supplied; ortherwise initialize as a vector of zeros
    if not isinstance(b_vec, np.ndarray):
        b_vec = np.zeros((n_rows, 1))
    a_mat = np.append(a_mat, b_vec, axis=1)
    n_cols += 1
    for i in range(n_cols):
        a_mat_max, max_idx = 0., piv_idx
        # locate the pivot index by searching each row for a non-zero value.
        for j in range(piv_idx, n_rows):
            # if a pivot index is found, set the value to the col index for the row in which it was found
            if a_mat[j, i] > a_mat_max:
                max_idx = j
                a_mat_max = a_mat[j, i]
            # if a pivot index is not found in a given row, reset a_mat_max to -1 and move to the next row
            elif j == n_rows-1 and a_mat_max == 0.:
                piv_idx = max_idx
                a_mat_max = -1.
        # update the matrix by flipping the row and columns to achieve row echelon form
        if a_mat_max > 0.:
            if max_idx > piv_idx:
                a_mat[[piv_idx, max_idx]] = a_mat[[max_idx, piv_idx]]
            for j in range(piv_idx + 1, n_rows):
                if a_mat[j, i] == 1.:
                    a_mat[j, i:n_cols] = np.fmod(a_mat[j, i:n_cols] + a_mat[piv_idx, i:n_cols], 2)
            piv_idx += 1
    # extract the solution from the bottom to the top since it is now in row echelon form
    b_vec = a_mat[0:n_rows, n_cols-1].tolist()
    for i in range(n_rows - 1, -1, -1):
        col_idx, z_free = -1., []
        for j in range(n_cols-1):
            if a_mat[i, j] == 1.:
                if col_idx == -1:
                    col_idx = j
                else:
                    z_free.append(j)
        if col_idx >= 0.:
            z_vals.append([col_idx, z_free, b_vec[i]])
    # check for free solutions -- select 0 for the free solution
    # for the ILC generator screening procedure, 0 leads to an I op and 1 leads to a Z Pauli op
    for z_val in (z_vals):
        b_val = z_val[2]
        for z_free in (z_val[1]):
            if z_sln[z_free] == -1:
                z_sln[z_free] = 0.
            b_val = np.fmod(b_val + z_sln[z_free], 2)
        z_sln[z_val[0]] = b_val
    # check that z_sln does not have any -1 values left -- if so, a solution was not found.
    for z_val in z_sln:
        if z_val == -1:
            warnings.warn("Gaussian elimination over GF(2) failed to find a solution.", RuntimeWarning)
    return np.array(z_sln)


def get_ilc_params_by_diag(qubit_ham, ilc_gens, qmf_var_params):
    """Driver function that solves the generalized eigenvalue problem Hc = ESc required
    to obtain the ground state coefficients (ILC parameters). These are subsequently recast
    according to Appendix C of Ref. 1 in a form that is suitable for constructing ILC circuits.

    Args:
        qubit_ham (QubitOperator): the qubit Hamiltonian of the system.
        ilc_gens (list of QubitOperator): the anticommuting set of ILC Pauli words.

    Returns:
        list of float: the ILC parameters corresponding to the ACS of ILC generators
    """

    # Add the identity operator to the local copy of the ACS
    ilc_gens.insert(0, QubitOperator.identity())
    n_var_params = len(ilc_gens)
    qubit_ham_mat = np.zeros((n_var_params, n_var_params), dtype=complex)
    qubit_overlap_mat = np.zeros((n_var_params, n_var_params), dtype=complex)

    # Construct the lower triangular matrices for the qubit Hamiltonian and overlap integrals
    for i in range(n_var_params):
        # H T_i|QMF> = H |psi_i>
        h_psi_i = qubit_ham * ilc_gens[i]

        # <QMF|T_i H T_i|QMF> = <psi_i| H | psi_i> = H_ii
        qubit_ham_mat[i, i] = get_op_expval(ilc_gens[i] * h_psi_i, qmf_var_params)

        # <QMF|T_i T_i|QMF> = <psi_i|psi_i> = 1
        qubit_overlap_mat[i, i] = 1. + 0j

        for j in range(i + 1, n_var_params):
            # <QMF|T_j H T_i|QMF> = <psi_j| H | psi_i> = H_ji
            qubit_ham_mat[j, i] = get_op_expval(ilc_gens[j] * h_psi_i, qmf_var_params)

            # <QMF|T_j T_i|QMF> = <psi_j|psi_i> --> exactly zero only for pure QMF states
            qubit_overlap_mat[j, i] = get_op_expval(ilc_gens[j] * ilc_gens[i], qmf_var_params)
            if i == 0:
                qubit_ham_mat[j, i] *= 1j
                qubit_overlap_mat[j, i] *= 1j

    # Solve the generalized eigenvalue problem
    _, subspace_coefs = scipy.linalg.eigh(a=qubit_ham_mat, b=qubit_overlap_mat, lower=True, driver="gvd")

    # Compute the ILC parameters using the ground state coefficients
    gs_coefs = subspace_coefs[:, 0]
    if gs_coefs[0].real > 0.:
        gs_coefs *= -1.
    denom_sum, ilc_var_params = 0., []
    for i in range(2):
        denom_sum += pow(gs_coefs[i].real, 2.) + pow(gs_coefs[i].imag, 2.)
    beta_1 = np.arcsin(gs_coefs[1] / np.sqrt(denom_sum))
    ilc_var_params.append(beta_1.real)
    for i in range(2, n_var_params):
        denom_sum += pow(gs_coefs[i].real, 2.) + pow(gs_coefs[i].imag, 2.)
        beta = np.arcsin(gs_coefs[i] / np.sqrt(denom_sum))
        ilc_var_params.append(beta.real)
    del ilc_gens[0]
    return ilc_var_params



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module that defines helper functions to taper qubits in a molecular problem
with Z2 tapering.

For all chemical Hamiltonians, there are at least two symmetries (electron
number and spin conservation) that reduce the qubits count by two. Furthermore,
molecular symmetries can lead to a reduction of the number of qubits required to
encode a problem. Those symmetries can be interpreted as degenerated
eigenvalues. In the real space, symmetry operations lead to the same total
energy.

Ref:
    Tapering off qubits to simulate fermionic Hamiltonians
    Sergey Bravyi, Jay M. Gambetta, Antonio Mezzacapo, Kristan Temme.
    arXiv:1701.08213
"""

from functools import reduce
import operator

import numpy as np

from tangelo.toolboxes.qubit_mappings.statevector_mapping import get_vector
from tangelo.toolboxes.operators.multiformoperator import MultiformOperator, ConvertPauli, do_commute


def get_z2_taper_function(unitary, kernel, q_indices, n_qubits, n_symmetries, eigenvalues=None):
    """Defines a method for applying taper to an arbitrary operator in this
    space. The operator is first conditioned against the tapering-operators:
    terms which do not commute with the Hamiltonian's symmetries are culled, and
    the remaining operator is rotated against the tapering unitary. After
    applying the eigenvalues to account for tapered elements, the operator is
    then finally tapered, with the relevant qubits removed.

    Args:
        unitary (array of float): Unitary matrix to perform U*HU.
        kernel (array of bool): Kernel representing the NULL space for the
            operator.
        q_indices (array of int): Indices for the relevant columns in the
            operator array representation.
        n_qubits (int): Self-explanatory.
        n_symmetries (int): Number of qubits to remove.
        eigenvalues (array of int): Initial state eigenvalues for the qubits
            to be removed.

    Returns:
        function: function for tapering operator.
    """

    def do_taper(operator, eigenvalues=eigenvalues):

        # Remove non-commuting terms.
        commutes = do_commute(operator, kernel, term_resolved=True)
        indices = np.where(commutes is False)[0]

        if len(indices) > 0:
            operator.remove_terms(indices)

        # Apply rotation if the operator is not trivial.
        if operator.n_terms == 0:
            op_matrix, factors = np.zeros(operator.n_qubits), np.array([0.0])
        else:
            product = operator * unitary
            product.compress()
            product_reverse = unitary * product
            product_reverse.compress()
            op_matrix, factors = product_reverse.integer, product_reverse.factors

        if factors.max() == 0.0:
            return MultiformOperator.from_integerop(np.zeros((1, n_qubits-n_symmetries), dtype=int), np.array([0.0]))

        for index, eigenvalue in zip(q_indices, eigenvalues):
            factors[op_matrix[:, index] > 0] *= eigenvalue

        tapered = np.delete(op_matrix, q_indices, axis=1)

        return MultiformOperator.from_integerop(tapered, factors)

    return do_taper


def get_clifford_operators(kernel):
    """Function to identify, with a kernel, suitable single-Pauli gates and the
    related unitary Clifford gates.

    Args:
        kernel (array-like of bool): Array of M x 2N booleans, where M is the
            number of terms and N is the number of qubits. Refers to a qubit
            operator in the stabilizer notation.

    Returns:
        (list of MultiformOperator, list of int): Encoded binary-encoded Pauli
            strings and symmetry indices.
    """

    indices = list()
    n_qubits = kernel.shape[1] // 2
    cliffords = list()
    factors = np.full(2, np.sqrt(0.5))

    for row, ki in enumerate(kernel):
        vector = np.zeros(n_qubits * 2, dtype=int)
        rest = np.delete(kernel, row, axis=0)

        for col in range(n_qubits):
            tau_i = ConvertPauli((ki[col], ki[col+n_qubits])).integer
            tau_j = [ConvertPauli((rj[col], rj[col+n_qubits])).integer for rj in rest]

            # Default value for the conversion (identity). Find a suitable
            # choice of Pauli gate which anti-commutes with one symmetry
            # operator and commutes with all others.
            pauli = (0, 0)
            for pauli_i in range(3):
                tau_destination = np.delete(np.array([1, 2, 3]), pauli_i)
                lookup = [0, pauli_i + 1]

                if all(oi in lookup for oi in tau_j):
                    if tau_i in tau_destination:
                        pauli = ConvertPauli(pauli_i + 1).tuple
                        break

            if sum(pauli) > 0:
                vector[[col, col + n_qubits]] = pauli
                indices.append(col)
                clifford = np.array([vector, ki])
                cliffords.append(MultiformOperator.from_binaryop(bin_op=clifford, factors=factors))
                break

    return cliffords, np.array(indices)


def get_unitary(cliffords):
    """Function generating the product over multiple Clifford
    operators as a single unitary. The result is a MultiformOperator.

    Args:
        cliffords (list of MultiformOperator): Encoded Clifford operators.

    Returns:
        MultiformOperator: Multiplication reflecting the composite operator.
    """
    return reduce(operator.mul, cliffords[1:], cliffords[0])


def get_eigenvalues(symmetries, n_qubits, n_electrons, spin, mapping, up_then_down):
    """Get the initial state eigenvalues, as operated on by each of the symmetry
    operators. These are used to capture the action of each Pauli string in the
    Hamiltonian on the tapered qubits.

    Args:
        symmetries (array-like of bool): Symmetries in binary encoding.
        n_qubits (int): Self-explanatory.
        n_electrons (int): Self-explanatory.
        mapping (str): Qubit mapping.
        up_then_down (bool): Whether or not spin ordering is all up then
            all down.

    Returns:
        array of +/-1: Eigenvalues of an operator with symmetries.
    """

    psi_init = get_vector(n_qubits, n_electrons, mapping, up_then_down, spin)

    if len(symmetries.shape) == 1:
        symmetries = np.reshape(symmetries, (-1, len(symmetries)))

    each_qubit = np.einsum("ij,j->ij", symmetries[:, n_qubits:].astype(bool), psi_init)
    eigenvalues = np.product(-2 * each_qubit + 1, axis=1)

    return eigenvalues



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import numpy as np

from tangelo.molecule_library import mol_H2_sto3g, mol_H4_cation_sto3g
from tangelo.toolboxes.qubit_mappings import jordan_wigner
from tangelo.toolboxes.ansatz_generator.uccgd import UCCGD
from tangelo.linq import Simulator


class UCCGDTest(unittest.TestCase):

    def test_uccgd_set_var_params(self):
        """Verify behavior of set_var_params for different inputs (keyword,
        list, numpy array).
        """

        uccgd_ansatz = UCCGD(mol_H2_sto3g)

        three_ones = np.ones((3,))

        uccgd_ansatz.set_var_params("ones")
        np.testing.assert_array_almost_equal(uccgd_ansatz.var_params, three_ones, decimal=6)

        uccgd_ansatz.set_var_params([1., 1., 1.])
        np.testing.assert_array_almost_equal(uccgd_ansatz.var_params, three_ones, decimal=6)

        uccgd_ansatz.set_var_params(np.array([1., 1., 1.]))
        np.testing.assert_array_almost_equal(uccgd_ansatz.var_params, three_ones, decimal=6)

    def test_uccgd_incorrect_number_var_params(self):
        """Return an error if user provide incorrect number of variational
        parameters.
        """

        upccgsd_ansatz = UCCGD(mol_H2_sto3g)

        self.assertRaises(ValueError, upccgsd_ansatz.set_var_params, np.array([1., 1., 1., 1.]))

    def test_uccgd_H2(self):
        """Verify closed-shell UCCGD functionalities for H2."""

        # Build circuit
        uccgd_ansatz = UCCGD(mol_H2_sto3g)
        uccgd_ansatz.build_circuit()

        # Build qubit hamiltonian for energy evaluation
        qubit_hamiltonian = jordan_wigner(mol_H2_sto3g.fermionic_hamiltonian)

        # Assert energy returned is as expected for given parameters
        sim = Simulator()
        uccgd_ansatz.update_var_params([0.78525105, 1.14993361, 1.57070471])
        energy = sim.get_expectation_value(qubit_hamiltonian, uccgd_ansatz.circuit)
        self.assertAlmostEqual(energy, -1.1372701, delta=1e-6)

    def test_uccgd_H4_open(self):
        """Verify open-shell UCCGD functionalities for H4."""

        var_params = [0.79092606, 0.47506062, 0.80546313, 0.88827205, 1.35659603,
                      0.85509053, 0.86380802, 0.89392798, 0.71922788, 1.36597650,
                      1.41426772, 0.96502797, 0.90349484, 0.58184862, 0.91536039,
                      0.64726991, 0.78419327, 1.59344719, 0.88236760, 1.26174588,
                      0.78342616, 1.21796644, 1.21234109, 0.29929539, 1.53074675,
                      1.41872463, 0.85843055, 1.10503577, 1.19597016, 0.92342705,
                      1.02231599]

        # Build circuit
        uccgd_ansatz = UCCGD(mol_H4_cation_sto3g)
        uccgd_ansatz.build_circuit()

        # Build qubit hamiltonian for energy evaluation
        qubit_hamiltonian = jordan_wigner(mol_H4_cation_sto3g.fermionic_hamiltonian)

        # Assert energy returned is as expected for given parameters
        sim = Simulator()
        uccgd_ansatz.update_var_params(var_params)
        energy = sim.get_expectation_value(qubit_hamiltonian, uccgd_ansatz.circuit)
        self.assertAlmostEqual(energy, -1.64190668, delta=1e-6)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .qubit_terms_grouping import group_qwc, exp_value_from_measurement_bases,\
    check_bases_commute_qwc, map_measurements_qwc
from .estimate_measurements import get_measurement_estimate
from .classical_shadows.classical_shadows import ClassicalShadow
from .classical_shadows.randomized import RandomizedClassicalShadow
from .classical_shadows.derandomized import DerandomizedClassicalShadow
from .classical_shadows.adaptive import AdaptiveClassicalShadow



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import numpy as np
import os
from openfermion import load_operator

from tangelo.molecule_library import mol_H2_sto3g, mol_H4_doublecation_minao
from tangelo.toolboxes.qubit_mappings import jordan_wigner
from tangelo.toolboxes.ansatz_generator.hea import HEA

from tangelo.linq import Simulator

# Initiate simulator
sim = Simulator()

# For openfermion.load_operator function.
pwd_this_test = os.path.dirname(os.path.abspath(__file__))


class HEATest(unittest.TestCase):

    def test_hea_set_var_params(self):
        """Verify behavior of set_var_params for different inputs (keyword,
        list, numpy array).
        """

        hea_ansatz = HEA(molecule=mol_H2_sto3g)

        hea_ansatz.set_var_params("ones")
        np.testing.assert_array_almost_equal(hea_ansatz.var_params, np.ones(4 * 3 * 3), decimal=6)

        hea_ansatz.set_var_params(np.ones(4 * 3 * 3))
        np.testing.assert_array_almost_equal(hea_ansatz.var_params, np.ones(4 * 3 * 3), decimal=6)

        hea_ansatz.set_var_params("zeros")
        np.testing.assert_array_almost_equal(hea_ansatz.var_params, np.zeros(4 * 3 * 3), decimal=6)

    def test_hea_incorrect_number_var_params(self):
        """Return an error if user provide incorrect number of variational
        parameters.
        """

        hea_ansatz = HEA(molecule=mol_H2_sto3g)

        self.assertRaises(ValueError, hea_ansatz.set_var_params, np.ones(4 * 3 * 3 + 1))

    def test_hea_H2(self):
        """Verify closed-shell HEA functionalities for H2."""

        # Build circuit
        hea_ansatz = HEA(molecule=mol_H2_sto3g)
        hea_ansatz.build_circuit()

        # Build qubit hamiltonian for energy evaluation
        qubit_hamiltonian = jordan_wigner(mol_H2_sto3g.fermionic_hamiltonian)

        params = [ 1.96262489e+00, -9.83505909e-01, -1.92659544e+00,  3.68638855e+00,
                   4.71410736e+00,  4.78247991e+00,  4.71258582e+00, -4.79077006e+00,
                   4.60613188e+00, -3.14130503e+00,  4.71232383e+00,  1.35715841e+00,
                   4.30998410e+00, -3.00415626e+00, -1.06784872e+00, -2.05119893e+00,
                   6.44114344e+00, -1.56358255e+00, -6.28254779e+00,  3.14118427e+00,
                  -3.10505551e+00,  3.15123780e+00, -3.64794717e+00,  1.09127829e+00,
                   4.67093656e-01, -1.19912860e-01,  3.99351728e-03, -1.57104046e+00,
                   1.56811666e+00, -3.14050540e+00,  4.71181097e+00,  1.57036595e+00,
                  -2.16414405e+00,  3.40295404e+00, -2.87986715e+00, -1.69054279e+00]

        # Assert energy returned is as expected for given parameters
        hea_ansatz.update_var_params(params)
        energy = sim.get_expectation_value(qubit_hamiltonian, hea_ansatz.circuit)
        self.assertAlmostEqual(energy, -1.1372661564779496, delta=1e-6)

    def test_hea_H4_doublecation(self):
        """Verify closed-shell HEA functionalities for H4 2+."""

        var_params = [-2.34142720e-04,  6.28472420e+00,  4.67668267e+00, -3.14063369e+00,
                      -1.53697174e+00, -6.22546556e+00, -3.11351342e+00,  3.14158366e+00,
                      -1.57812617e+00,  3.14254101e+00, -6.29656957e+00, -8.93646210e+00,
                       7.84465163e-04, -1.32569792e-05,  1.90710480e+00,  6.44924149e+00,
                      -4.69797158e+00,  3.47688734e+00,  8.24677008e-04,  1.89006312e-04,
                      -1.57734737e+00, -3.63191375e+00, -3.15706332e+00, -1.73625091e+00]

        # Build circuit with Ry rotationas instead of RZ*RX*RZ rotations
        hea_ansatz = HEA(molecule=mol_H4_doublecation_minao, rot_type="real")
        hea_ansatz.build_circuit()

        # Build qubit hamiltonian for energy evaluation
        qubit_hamiltonian = load_operator("mol_H4_doublecation_minao_qubitham_jw.data", data_directory=pwd_this_test+"/data", plain_text=True)

        # Assert energy returned is as expected for given parameters
        hea_ansatz.update_var_params(var_params)
        energy = sim.get_expectation_value(qubit_hamiltonian, hea_ansatz.circuit)
        self.assertAlmostEqual(energy, -0.795317, delta=1e-4)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Define an abstract quantum gate class holding information about a quantum
gate operation, without tying it to a particular backend or an underlying
mathematical operation.
"""
from math import pi
from typing import Union

from numpy import integer, ndarray, floating

ONE_QUBIT_GATES = {"H", "X", "Y", "Z", "S", "T", "RX", "RY", "RZ", "PHASE"}
TWO_QUBIT_GATES = {"CNOT", "CX", "CY", "CZ", "CRX", "CRY", "CRZ", "CPHASE", "XX", "SWAP"}
THREE_QUBIT_GATES = {"CSWAP"}

ONE_TARGET_GATES = {"H", "X", "Y", "Z", "S", "T", "RX", "RY", "RZ", "PHASE",
                    "CNOT", "CX", "CY", "CZ", "CRX", "CRY", "CRZ", "CPHASE"}
TWO_TARGET_GATES = {"XX", "SWAP", "CSWAP"}

PARAMETERIZED_GATES = {"RX", "RY", "RZ", "PHASE", "CRX", "CRY", "CRZ", "CPHASE", "XX"}
INVERTIBLE_GATES = {"H", "X", "Y", "Z", "S", "T", "RX", "RY", "RZ", "CH", "PHASE",
                    "CNOT", "CX", "CY", "CZ", "CRX", "CRY", "CRZ", "CPHASE", "XX", "SWAP"
                    "CSWAP"}


class Gate(dict):
    """An abstract gate class that exposes all the gate information such as gate
    name, control and target qubit indices, parameter values. Assumes qubit
    indexation starts at index 0.

    Attributes:
        name (str): the gate name,
        target (int): A positive integer denoting the index of the target qubit,
        control (int): A positive integer denoting the index of the control
            qubit,
        parameter: A floating-point number or symbolic expression that will
            resolve at runtime. Can hold anything.
        is_variational (bool): a boolean indicating if the gate operation is
            variational or not.
    """

    def __init__(self, name: str, target: Union[int, integer, list, ndarray],
                 control: Union[int, integer, list, ndarray] = None,
                 parameter="", is_variational: bool = False):
        """ This gate class is basically a dictionary with extra methods. """

        def check_qubit_indices(qubit_list, label):
            """Function to check if all given qubit indices are positive integers

            Args:
                qubit_list (list) :: List of values to check are positive integers
                label (str) :: The label of the list, "control" or "target"

            Raises:
                ValueError :: If any of the values in the list are not non-negative integers.
            """
            errmsg = ""
            for ind in qubit_list:
                if (type(ind) != int) or (ind < 0):
                    errmsg += f"\n {label} qubit index {ind} is not a non-negative integer"

            if errmsg:
                raise ValueError(f"Error: type or value of {label} qubit indices not as expected. See details below:"+errmsg)

        target = (target.tolist() if isinstance(target, ndarray) else list(target)) if hasattr(target, "__iter__") else [target]

        check_qubit_indices(target, "target")

        if control is not None:
            if name[0] != "C":
                raise ValueError(f"Gate {name} was given control={control} but does not support controls. Try C{name}")
            control = (control.tolist() if isinstance(control, ndarray) else list(control)) if hasattr(control, "__iter__") else [control]
            check_qubit_indices(control, "control")

        # Check for duplications in qubits
        all_involved_qubits = target if control is None else target + control
        if len(all_involved_qubits) != len(set(all_involved_qubits)):
            raise ValueError(f"There are duplicate qubits in the target/control qubits")

        # Check for the number of qubits. If it is a custom gate (not defined in
        # the X_QUBIT_GATES sets), ignore the check. We also do not prevent
        # the creation of multi-controlled gates (begins with "C").
        if name.upper() in ONE_TARGET_GATES:
            n_targets = 1
        elif name.upper() in TWO_TARGET_GATES:
            n_targets = 2
        else:
            n_targets = len(target)
        if len(target) != n_targets:
            raise ValueError(f"Gate {name}: expected {n_targets} target qubits, but got {target}")

        self.__dict__ = {"name": name, "target": target, "control": control,
                         "parameter": parameter, "is_variational": is_variational}

    def __str__(self):
        """Print gate information in a somewhat formatted way. Do not print
        empty attributes.
        """

        mystr = f"{self.name:<10}"
        for attr in ["target", "control"]:
            if self.__getattribute__(attr) or isinstance(self.__getattribute__(attr), int):
                mystr += f"{attr} : {self.__getattribute__(attr)}   "
        if self.__getattribute__("parameter") != "":
            mystr += f"parameter : {self.__getattribute__('parameter')}"
        if self.is_variational:
            mystr += "\t (variational)"

        return mystr

    def __eq__(self, other):
        """Define equality (==) operator on gates"""

        ds, do = self.__dict__, other.__dict__

        if any(ds[k] != do[k] for k in ds if k != "parameter"):
            return False

        parameter = round(ds["parameter"] % (2 * pi), 7) if isinstance(ds["parameter"], (float, int)) else ds["parameter"]
        other_parameter = round(do["parameter"] % (2 * pi), 7) if isinstance(do["parameter"], (float, int)) else do["parameter"]

        return parameter == other_parameter

    def __ne__(self, other):
        """Define inequality (!=) operator on gates"""
        return not (self == other)

    def inverse(self):
        """Return the inverse (adjoint) of a gate.

        Return:
            Gate: the inverse of the gate.
        """
        if self.name not in INVERTIBLE_GATES:
            raise AttributeError(f"{self.name} is not an invertible gate")
        if self.parameter == "":
            new_parameter = ""
        elif isinstance(self.parameter, (float, floating, int, integer)):
            new_parameter = -self.parameter
        elif self.name in {"T", "S"}:
            new_parameter = -pi / 2 if self.name == "T" else -pi / 4
            return Gate("PHASE", self.target, self.control, new_parameter, self.is_variational)
        else:
            raise AttributeError(f"{self.name} is not an invertible gate when parameter is {self.parameter}")
        return Gate(self.name, self.target, self.control, new_parameter, self.is_variational)

    def serialize(self):
        return {"type": "Gate",
                "params": {"name": self.name, "target": self.target,
                           "control": self.control, "parameter": self.parameter}}



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module defining the MultiformOperator class. It stores a qubit operator in
many forms: tuples, numpy array of int and stabilizer formalism. Most of the
internal methods use the numpy array of int. The main application for this class
is to identify commutation relation faster with the stabilizer notation.
"""

from operator import itemgetter

import numpy as np

from tangelo.helpers.math import bool_col_echelon
from tangelo.toolboxes.operators import QubitOperator, count_qubits


class MultiformOperator(QubitOperator):
    """Construct integer and binary representations of the operator, as based on
    the user-specified input. Internal operations are mostly done with an
    integer array. The conversion is defined as:
    Letter | Integer | Binary
    -    I   |    0    | (0,0)
    -    Z   |    1    | (0,1)
    -    X   |    2    | (1,0)
    -    Y   |    3    | (1,1)
    Most of the algorithms use the integer representation to detect symmetries
    and performed operation (ex: multiplication). At the end, the operator can
    be output as a QubitOperator.

    Attributes:
        terms (dict): Qubit terms.
        n_qubits (int): Number of qubits for this operator.
        factors (array-like of complex): Factors for each term. Without that,
            information would be lost when converting from a QubitOperator.
        integer (array-like of np.int8): Array of 0s, 1s, 2s and 3s. Each line
            represents a term, and each term is a Pauli word.
        binary (array-like of bool): Array of 0s (false) and 1s (true). Binary
            representation of the integer array.
        binary_swap (array-like of bool): Copy of self.binary but with swapped
            columns.
        kernel (array-like): Null space of the binary representation + identity.

    Properties:
        n_terms (int): Number of terms in this qubit Hamiltonian.
        qubit_operator (QubitOperator): Self-explanatory.
    """

    def __init__(self, terms, n_qubits, factors, integer, binary):

        # Parent class definition.
        super(QubitOperator, self).__init__()
        self.terms = terms

        self.n_qubits = n_qubits

        self.factors = factors
        self.integer = integer
        self.binary = binary

        # Swapping the binary representation. Example: instead of (0,1), write
        # (1,0). This is useful to detect commutation between terms.
        list_col_swap = list(range(n_qubits, 2*n_qubits)) + list(range(n_qubits))
        self.binary_swap = self.binary[:, list_col_swap]

        # Attribute to be defined when symmetries will be computed.
        self.kernel = None

    @property
    def n_terms(self):
        return len(self.terms)

    @property
    def qubitoperator(self):
        qubit_op = QubitOperator()
        qubit_op.terms = self.terms.copy()
        return qubit_op

    @classmethod
    def from_qubitop(cls, qubit_op, n_qubits=None):
        """Initialize MultiformOperator from a qubit operator."""

        if n_qubits is None:
            n_qubits = count_qubits(qubit_op)
        factors = np.array(list(qubit_op.terms.values()))
        int_op = qubit_to_integer(qubit_op, n_qubits)
        bin_op = integer_to_binary(int_op)

        return cls(qubit_op.terms, n_qubits, factors, int_op, bin_op)

    @classmethod
    def from_integerop(cls, int_op, factors):
        """Initialize MultiformOperator from an integer operator."""

        assert len(factors) == int_op.shape[0], \
             f"The number of factors ({len(factors)}) must be the same as the number of terms ({int_op.shape[0]})."
        terms = integer_to_qubit_terms(int_op, factors)
        bin_op = integer_to_binary(int_op)

        return cls(terms, int_op.shape[1], factors, int_op.astype(np.int8), bin_op)

    @classmethod
    def from_binaryop(cls, bin_op, factors):
        """Initialize MultiformOperator from a binary operator."""

        assert len(factors) == bin_op.shape[0], \
            f"The number of factors ({len(factors)}) must be the same as the number of terms ({bin_op.shape[0]})."
        n_qubits = bin_op.shape[1] // 2
        # The integer array is defined trivially with the binary array.
        int_op = 2 * bin_op[:, :n_qubits].astype(np.int8) + bin_op[:, n_qubits:].astype(np.int8)
        terms = integer_to_qubit_terms(int_op, factors)

        return cls(terms, n_qubits, factors, int_op, bin_op.astype(bool))

    def __mul__(self, other_operator):
        """Multiply two MultiformOperators together, return a MultiformOperator
        corresponding to the product.

        Args:
            other_operator (MultiformOperator): Another operator to multiply self.

        Returns:
            MultiformOperator: Product of the multiplication.
        """

        # Take into account the order of Pauli matrices multiplication.
        c_calc = np.array([[1, 1, 1, 1],
                           [1, 1, 1j, -1j],
                           [1, -1j, 1, 1j],
                           [1, 1j, -1j, 1]], dtype=complex)

        factors = np.zeros((self.n_terms * other_operator.n_terms), dtype=complex)
        product = np.zeros((factors.shape[0], self.n_qubits), dtype=int)
        increment = other_operator.n_terms

        for term_i, integer in enumerate(self.integer):
            new_cs = c_calc[self.integer[term_i], other_operator.integer]
            product[term_i * increment: (term_i + 1) * increment] = integer ^ other_operator.integer
            factors[term_i * increment: (term_i + 1) * increment] = self.factors[term_i] * other_operator.factors * np.product(new_cs, axis=1)

        product, factors = MultiformOperator.collapse(product, factors)

        return MultiformOperator.from_integerop(product, factors=factors)

    def get_kernel(self):
        """Get the kernel for a matrix of binary integers. The identity matrix
        is appended, and the extended matrix is then reduced to column-echelon
        form. The null space is then extracted.

        Returns:
            Array-like: 2d numpy array of binary integers.
        """

        identity_to_append = np.identity(2*self.n_qubits)

        # Append identity to the matrix E.
        E_prime = np.concatenate((self.binary, identity_to_append), axis=0)

        # Put the matrix into column-echelon form.
        E_prime = bool_col_echelon(E_prime.astype(bool))

        # Extracting the kernel.
        kernel = np.array([E_prime[self.binary.shape[0]:, i] for i in range(self.n_qubits)
                           if not E_prime[:self.binary.shape[0], i].max()])
        kernel = np.concatenate((kernel[:, self.n_qubits:], kernel[:, : self.n_qubits]), axis=1)

        self.kernel = kernel
        return kernel

    def remove_terms(self, indices):
        """Remove term(s) from operator. The associated entry is deleted from
        factors, integer, binary and binary_swap.

        Args:
            indices (int or list of int): Remove all the terms corresponding to
                the indices.
        """

        if isinstance(indices, int):
            indices = np.array([indices])

        self.factors = np.delete(self.factors, indices, axis=0)
        self.integer = np.delete(self.integer, indices, axis=0)
        self.binary = np.delete(self.binary, indices, axis=0)
        self.binary_swap = np.delete(self.binary_swap, indices, axis=0)

        # Update the terms attribute.
        self.terms = integer_to_qubit_terms(self.integer, self.factors)

    def _update(self):
        """Updates attributes according to self.terms."""

        # Updating attributes.
        self.n_qubits = count_qubits(self)
        self.factors = np.array([coeff for coeff in self.terms.values()])
        self.integer = qubit_to_integer(self, self.n_qubits)
        self.binary = integer_to_binary(self.integer)

        # Updating the binary_swap from the new binary.
        list_col_swap = list(range(self.n_qubits, 2*self.n_qubits)) + list(range(self.n_qubits))
        self.binary_swap = self.binary[:, list_col_swap]

        # Resetting the kernel attribute.
        self.kernel = None

    def compress(self, abs_tol=None):
        """Overloads the QubitOperator.compress method. It adds an update for
            the attributes.

        Args:
            abs_tol (float): Tolerance for the coefficients to be discarded or
                not. By default, it is EQ_TOLERANCE=1e-8 (in openfermion).
        """

        # If an update is done to EQ_TOLERANCE in openfermion, it will be
        # consistent in this implementation.
        if abs_tol is None:
            super(QubitOperator, self).compress()
        else:
            super(QubitOperator, self).compress(abs_tol)

        self._update()

    @staticmethod
    def collapse(operator, factors):
        """Function to identify and sum over duplicate terms in an operator, to
        collapse a set of Pauli words to their minimal representation.

        Args:
            operator (array of int): Operator in integer notation.
            factors (array of complex): Self-explanatory.

        Returns:
            (array of int, arrays of float): Array of unique integer-encoded
                Pauli words, their factors in the operator.
        """

        all_terms = np.concatenate((operator, np.linspace(0, len(operator) - 1, len(operator), dtype=int).reshape(len(operator), -1)), axis=1)

        qubits = np.linspace(0, operator.shape[1] - 1, operator.shape[1], dtype=int)

        sorted_terms = np.array(sorted(all_terms, key=itemgetter(*qubits)))
        sorted_factors = factors[sorted_terms[:, -1]]

        unique, inverse = np.unique(sorted_terms[:, :-1], axis=0, return_inverse=True)

        factors = np.zeros(len(unique), dtype=complex)

        for index in range(len(sorted_terms)):
            factors[inverse[index]] += sorted_factors[index]

        nonzero = np.where(abs(factors) > 0)
        unique = unique[nonzero]
        factors = factors[nonzero]

        if len(np.shape(unique)) == 1:
            unique = np.reshape(unique, (-1, len(unique)))

        return unique.astype(np.int8), factors


class ConvertPauli:
    """Helper class to help convert from/to string <-> int <-> stabilizer. It
    aims to replace multiple dictionaries. It performs the conversion of
        I <-> 0 <-> (0, 0)
        Z <-> 1 <-> (0, 1)
        X <-> 2 <-> (1, 0)
        Y <-> 3 <-> (1, 1)
    The input can be anything found in the previous conversion table.

    Args:
        pauli_id (string, int or tuple of bool): Single Pauli.

    Attributes:
        char (string): Pauli letter.
        integer (int): Pauli integer.
        tuple (tuple of bool): Pauli binary representation.
    """

    def __init__(self, pauli_id):
        pauli_translation = [
            ["I", 0, (0, 0)],
            ["Z", 1, (0, 1)],
            ["X", 2, (1, 0)],
            ["Y", 3, (1, 1)]
        ]

        self.char = None
        self.integer = None
        self.tuple = None

        # Going through the equivalences. When an input is found, attributes are
        # defined.
        for equiv in pauli_translation:
            if pauli_id in equiv:
                self.char, self.integer, self.tuple = equiv[:]

        # When the input is wrong (ex: "x", "G", (3,4), 42, ...), it raises an error.
        if self.char is None:
            raise ValueError(f"{pauli_id} is an unknown Pauli id.")


def qubit_to_integer(qubit_op, n_qubits=None):
    """Perform conversion from qubit operator to integer array. The integer
    attribute is instantiated, and populated from the QubitOperator attribute.

    Args:
        QubitOperator: self-explanatory.

    Returns:
        array-like of int: 2-D numpy array of 0s (I), 1s (Z), 2s (X) and 3s (Y)
            representing a qubit operator (shape[0] = number of terms, shape[1]
            = number of qubits.
    """

    if n_qubits is None:
        n_qubits = count_qubits(qubit_op)
    integer = np.zeros((len(qubit_op.terms), n_qubits), dtype=np.int8)

    for index, term in enumerate(qubit_op.terms):
        for n_qubit, pauli_letter in term:
            integer[index, n_qubit] = ConvertPauli(pauli_letter).integer

    return integer


def integer_to_binary(integer_op):
    """Perform conversion from integer array to binary (stabilizer) array.

    Args:
        interger_op (array-like of int): 2-D numpy array of 0s (I), 1s (Z), 2s
            (X) and 3s (Y) representing a qubit operator (shape[0] = number of
            terms, shape[1] = number of qubits.

    Returns:
        array-like of bool: Array of 0s and 1s representing the operator with
            the stabilizer notation.
    """

    n_terms = integer_op.shape[0]
    n_qubits = integer_op.shape[1]

    binary = np.zeros((n_terms, 2*n_qubits), dtype=bool)

    binary_x = (integer_op >> 1).astype(bool)
    binary_z = np.mod(integer_op, 2).astype(bool)

    np.concatenate((binary_x, binary_z), axis=1, out=binary)

    return binary


def integer_to_qubit_terms(integer_op, factors):
    """Perform conversion from integer array to qubit terms.

    Args:
        interger_op (array-like of int): 2-D numpy array of 0s (I), 1s (Z), 2s
            (X) and 3s (Y) representing a qubit operator (shape[0] = number of
            terms, shape[1] = number of qubits.
        factors (array-like of float): Coefficients for the qubit terms.

    Returns:
        dict: Pauli terms and coefficients.
    """

    qubit_operator = QubitOperator()
    for n_term, term in enumerate(integer_op):
        # Convert integer to a qubit terms (e.g. [0 1 2 3] => ((1, "Z"),
        # (2, "X"), (3, "Y"))).
        tuple_term = tuple([(qubit_i, ConvertPauli(int(term[qubit_i])).char) for
                             qubit_i in range(len(term)) if int(term[qubit_i]) > 0])
        # If a term is the same, QubitOperator takes into account the summation.
        qubit_operator += QubitOperator(tuple_term, factors[n_term])

    return qubit_operator.terms


def do_commute(hybrid_op_a, hybrid_op_b, term_resolved=False):
    """Check if two operators commute, using the stabilizer notation. With the
    *term_resolved* flag, the user can identify which terms in the parent
    operator do or do not commute with the target operator.

    In stabilizer notation, (ax|az)(bz|bx) = -1^(ax.bz + az.bx) (bz|bx)(ax|az)
    define the commutation relations. By evaluating the sign of (ax.bz + az.bx),
    we recover commutation relation. For a given pair of terms in the operators,
    ax.bz + az.bx = sum(XOR (AND (a' . b))) where a' is the z-x swapped
    a-vector. We then apply an OR reduction over all of the b-terms to
    identify whether term a_i commutes with all of b_j. Applying once again,
    an OR reduction over all a_i, we get the commutator for the entirety of
    a,b.

    Args:
        hybrid_op_a: First MultiformOperator.
        hybrid_op_b: Second MultiformOperator.
        term_resolved (bool): If True, get commutator for each term.

    Returns:
        bool or array of bool: If the operators commute or array of bool
            describing which terms are commuting.
    """

    term_bool = np.zeros(hybrid_op_a.n_terms, dtype=bool)

    # Binary_swap is used to check the commuation relation.
    for index, term in enumerate(hybrid_op_a.binary_swap):
        term_bool[index] = np.logical_or.reduce(
            np.logical_xor.reduce(term & hybrid_op_b.binary, axis=1)
            )

    if not term_resolved:
        return not np.all(term_bool)
    else:
        return np.logical_not(term_bool)



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import numpy as np

from tangelo.toolboxes.ansatz_generator.rucc import RUCC


class RUCCTest(unittest.TestCase):

    def test_construction_rucc(self):
        """Verify behavior of UCC1 and UCC3 construction. Those ansatze are
        constant (they do not change with the system because they always
        represent 4 spin-orbitals).
        """

        ucc1_ansatz = RUCC(1)
        ucc1_ansatz.build_circuit()
        assert(ucc1_ansatz.circuit.counts == {"X": 2, "RX": 2, "H": 6, "CNOT": 6, "RZ": 1})

        ucc3_ansatz = RUCC(3)
        ucc3_ansatz.build_circuit()
        assert(ucc3_ansatz.circuit.counts == {"X": 2, "RX": 4, "H": 6, "CNOT": 8, "RZ": 3})

    def test_ucc1_set_var_params(self):
        """Verify behavior of UCC1 set_var_params for different inputs (keyword,
        list, numpy array).
        """

        ucc1_ansatz = RUCC(1)

        ucc1_ansatz.set_var_params("zeros")
        np.testing.assert_array_almost_equal(ucc1_ansatz.var_params, np.array([0.]), decimal=6)

        ucc1_ansatz.set_var_params("ones")
        np.testing.assert_array_almost_equal(ucc1_ansatz.var_params, np.array([1.]), decimal=6)

        ucc1_ansatz.set_var_params([1.])
        np.testing.assert_array_almost_equal(ucc1_ansatz.var_params, np.array([1.]), decimal=6)

        ucc1_ansatz.set_var_params(np.array([1.]))
        np.testing.assert_array_almost_equal(ucc1_ansatz.var_params, np.array([1.]), decimal=6)

    def test_ucc3_set_var_params(self):
        """Verify behavior of UCC3 set_var_params for different inputs (keyword,
        list, numpy array).
        """

        ucc3_ansatz = RUCC(3)

        ucc3_ansatz.set_var_params("zeros")
        np.testing.assert_array_almost_equal(ucc3_ansatz.var_params, np.array([0., 0., 0.]), decimal=6)

        ucc3_ansatz.set_var_params("ones")
        np.testing.assert_array_almost_equal(ucc3_ansatz.var_params, np.array([1., 1., 1.]), decimal=6)

        ucc3_ansatz.set_var_params([1., 1., 1.])
        np.testing.assert_array_almost_equal(ucc3_ansatz.var_params, np.array([1., 1., 1.]), decimal=6)

        ucc3_ansatz.set_var_params(np.array([1., 1., 1.]))
        np.testing.assert_array_almost_equal(ucc3_ansatz.var_params, np.array([1., 1., 1.]), decimal=6)

    def test_rucc_wrong_n_params(self):
        """Verify RUCC wrong number of parameters."""

        with self.assertRaises(ValueError):
            RUCC(n_var_params=999)

        with self.assertRaises(ValueError):
            RUCC(n_var_params="3")

        with self.assertRaises(ValueError):
            RUCC(n_var_params=3.141516)

        with self.assertRaises(AssertionError):
            ucc3 = RUCC(n_var_params=3)
            ucc3.build_circuit()
            ucc3.update_var_params([3.1415])


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Jordan-Wigner transform on fermionic operators."""

from openfermion.transforms import jordan_wigner as openfermion_jordan_wigner


def jordan_wigner(operator):
    r"""Apply the Jordan-Wigner transform to a FermionOperator,
    InteractionOperator, or DiagonalCoulombHamiltonian to convert to a
    QubitOperator.

    Operators are mapped as follows:
    a_j^\dagger -> Z_0 .. Z_{j-1} (X_j - iY_j) / 2
    a_j -> Z_0 .. Z_{j-1} (X_j + iY_j) / 2

    Returns:
        QubitOperator: An instance of the QubitOperator class.

    Warning:
        The runtime of this method is exponential in the maximum locality of the
        original FermionOperator.

    Raises:
        TypeError: Operator must be a FermionOperator,
        DiagonalCoulombHamiltonian, or InteractionOperator.
    """
    qubit_operator = openfermion_jordan_wigner(operator)

    return qubit_operator
